/*****************************************************************************/
/* Include Header File.                                                      */
/*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <fcntl.h>
#include <unistd.h>
#include <termios.h>
#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/keysym.h>
#include <X11/Xaw/Label.h>

/*****************************************************************************/
/* Parameters.                                                               */
/*****************************************************************************/

#define DEFAULT_WIDTH     400
#define DEFAULT_HEIGHT    300
#define DEFAULT_FONTNAME "8x16"
#define DEFAULT_FONT_WIDTH   8
#define DEFAULT_FONT_HEIGHT  16
#define DEFAULT_PANEL_WIDTH   (8 * 80)
#define DEFAULT_PANEL_HEIGHT  (16 * 5)
#define DEFAULT_THICKNESS 5
#define DEFAULT_WAIT 100000

/*****************************************************************************/
/* Static Data.                                                              */
/*****************************************************************************/

typedef struct struct_viewer {
  XtAppContext app_context;
  Display * display;
  int screen;
  Colormap colormap;
  Widget toplevel;
  Window root_window, toplevel_window, canvas_window, panel_window;

  Widget ls_label, ls_come, ls_pass;
  Widget rs_label, rs_come, rs_pass;
  Widget intr;
  Widget init_button, quit_button;
  Widget upsw, downsw, limit, light;

  Pixmap canvas_pixmap;
  Atom atom_1, atom_2;
  unsigned long int foreground_color;
  unsigned long int background_color;
  GC gc_foreground, gc_background, gc_copy;
  GC gc_line, gc_line_clear, gc_line_green, gc_line_red;
  char * fontname;
  Font font;
  XFontStruct * font_struct;
  int x, y, width, height, depth;
  int thickness, wait;

  struct {
    int ls_come, ls_pass;
    int rs_come, rs_pass;
    int intr;
  } status;
  struct {
    int fd;
    int bar_angle;
    int bar_angle_delta;
    int upsw;
    int downsw;
    int limit;
    int light;
  } params;
} VIEWER;

/*****************************************************************************/
/* Set Default Values.                                                       */
/*****************************************************************************/

static void InitializeValues(VIEWER * wp)
{
  memset(wp, 0, sizeof(*wp));
  wp->width     = DEFAULT_WIDTH;
  wp->height    = DEFAULT_HEIGHT;
  wp->thickness = DEFAULT_THICKNESS;
  wp->wait      = DEFAULT_WAIT;
  wp->fontname  = DEFAULT_FONTNAME;
}

/*****************************************************************************/
/* Initializing Display.                                                     */
/*****************************************************************************/

static void CreateCanvas(VIEWER * wp, int argc, char ** argv)
{
  int n;
  Arg args[10];

  n = 0;
  XtSetArg(args[n], XtNwidth , wp->width ); n++;
  XtSetArg(args[n], XtNheight, wp->height); n++;
  wp->toplevel = XtAppInitialize(&(wp->app_context), "Viewer", NULL, 0,
				 &argc, argv, NULL, args, n);
  wp->display = XtDisplay(wp->toplevel);
  wp->screen = DefaultScreen(wp->display);
  wp->colormap = DefaultColormap(wp->display, wp->screen);
  wp->depth = DefaultDepth(wp->display, wp->screen);
  wp->root_window = XRootWindow(wp->display, wp->screen);

  XtRealizeWidget(wp->toplevel);
  wp->toplevel_window = XtWindow(wp->toplevel);

  n = 0;
  XtSetArg(args[n], XtNwidth , &(wp->width )); n++;
  XtSetArg(args[n], XtNheight, &(wp->height)); n++;
  XtSetArg(args[n], XtNforeground, &(wp->foreground_color)); n++;
  XtSetArg(args[n], XtNbackground, &(wp->background_color)); n++;
  XtGetValues(wp->toplevel, args, n);

  wp->canvas_window =
    XCreateSimpleWindow(wp->display, wp->toplevel_window,
			0, 0, wp->width, wp->height, 0,
			wp->foreground_color, wp->background_color);

  wp->canvas_pixmap = XCreatePixmap(wp->display, wp->canvas_window,
				    wp->width, wp->height, wp->depth);

  wp->font = XLoadFont(wp->display, wp->fontname);
  wp->font_struct = XQueryFont(wp->display, wp->font);  
}

static Widget CreatePanelButton(VIEWER * wp, int x, int y, int w, int h, int b)
{
  int n;
  Arg args[10];
  Widget wid;

  n = 0;
  XtSetArg(args[n], XtNx, x * DEFAULT_FONT_WIDTH); n++;
  XtSetArg(args[n], XtNy, y * DEFAULT_FONT_HEIGHT); n++;
  XtSetArg(args[n], XtNwidth, w * DEFAULT_FONT_WIDTH); n++;
  XtSetArg(args[n], XtNheight, h * DEFAULT_FONT_HEIGHT); n++;
  XtSetArg(args[n], XtNborder, b); n++;
  wid = XtCreateWidget("", labelWidgetClass, wp->toplevel, args, n);
  XtRealizeWidget(wid);
  return (wid);
}

static void CreatePanel(VIEWER * wp)
{
/*
  01234567890123456789012345678901234567890123456789012345678901234567890123
  LEFTSENSOR  COME PASS
  RIGHTSENSOR COME PASS
  INTR INIT QUIT

  UPSW DOWNSW LIMIT LIGHT
  */
  /*                                       x   y   w  h  b */
  wp->ls_label    = CreatePanelButton(wp,  0,  0, 10, 1, 0);
  wp->rs_label    = CreatePanelButton(wp,  0,  1, 11, 1, 0);
  wp->ls_come     = CreatePanelButton(wp, 12,  0,  4, 1, 1);
  wp->rs_come     = CreatePanelButton(wp, 12,  1,  4, 1, 1);
  wp->ls_pass     = CreatePanelButton(wp, 17,  0,  4, 1, 1);
  wp->rs_pass     = CreatePanelButton(wp, 17,  1,  4, 1, 1);

  wp->intr        = CreatePanelButton(wp,  0,  2,  4, 1, 1);
  wp->init_button = CreatePanelButton(wp,  5,  2,  4, 1, 1);
  wp->quit_button = CreatePanelButton(wp, 10,  2,  4, 1, 1);

  wp->upsw   = CreatePanelButton(wp,  0,  4,  4, 1, 1);
  wp->downsw = CreatePanelButton(wp,  5,  4,  6, 1, 1);
  wp->limit  = CreatePanelButton(wp, 12,  4,  5, 1, 1);
  wp->light  = CreatePanelButton(wp, 18,  4,  5, 1, 1);
}

static void CreateGC(VIEWER * wp)
{
  unsigned long int black;

  black = BlackPixel(wp->display, wp->screen);
  wp->gc_foreground = XCreateGC(wp->display, wp->canvas_window, 0, 0);
  wp->gc_background = XCreateGC(wp->display, wp->canvas_window, 0, 0);

  XSetForeground(wp->display, wp->gc_foreground, wp->foreground_color);
  XSetBackground(wp->display, wp->gc_foreground, wp->background_color);
  XSetForeground(wp->display, wp->gc_background, wp->background_color);
  XSetBackground(wp->display, wp->gc_background, wp->foreground_color);

  wp->gc_copy = XCreateGC(wp->display, wp->canvas_window, 0, 0);
  XSetFunction(wp->display, wp->gc_copy, GXcopy);

  XSetFont(wp->display, wp->gc_foreground, wp->font);
  XSetFont(wp->display, wp->gc_background, wp->font);
}

static void DrawPanelString(VIEWER * wp, Widget w, char * str, int flag)
{
  XDrawImageString(wp->display, XtWindow(w),
		   flag ? wp->gc_foreground : wp->gc_background,
		   0, wp->font_struct->ascent, str, strlen(str));
}

#if 0
static void DrawPanelIntValue(VIEWER * wp, Widget w, int a)
{
  char buf[100];
  sprintf(buf, "%d", a);
  XDrawImageString(wp->display, XtWindow(w), wp->gc_foreground,
		   0, wp->font_struct->ascent, buf, strlen(buf));
}

static void DrawPanelDoubleValue(VIEWER * wp, Widget w, double a)
{
  char buf[100];
  sprintf(buf, "%lf", a);
  XDrawImageString(wp->display, XtWindow(w), wp->gc_foreground,
		   0, wp->font_struct->ascent, buf, strlen(buf));
}
#endif

static void DrawPanel(VIEWER * wp)
{
  DrawPanelString(wp, wp->ls_label, "LEFTSENSOR" , 1);
  DrawPanelString(wp, wp->ls_come , "COME"       , !wp->status.ls_come);
  DrawPanelString(wp, wp->ls_pass , "PASS"       , !wp->status.ls_pass);

  DrawPanelString(wp, wp->rs_label, "RIGHTSENSOR", 1);
  DrawPanelString(wp, wp->rs_come , "COME"       , !wp->status.rs_come);
  DrawPanelString(wp, wp->rs_pass , "PASS"       , !wp->status.rs_pass);

  DrawPanelString(wp, wp->intr,        "INTR" , !wp->status.intr);
  DrawPanelString(wp, wp->init_button, "INIT" , 0);
  DrawPanelString(wp, wp->quit_button, "QUIT" , 0);

  DrawPanelString(wp, wp->upsw,   "UPSW"  , !wp->params.upsw);
  DrawPanelString(wp, wp->downsw, "DOWNSW", !wp->params.downsw);
  DrawPanelString(wp, wp->limit,  "LIMIT" , !wp->params.limit);
  DrawPanelString(wp, wp->light,  "LIGHT" , !wp->params.light);
}

static void InitializeDisplay(VIEWER * wp, int argc, char ** argv)
{
  CreateCanvas(wp, argc, argv);
  CreatePanel(wp);
  CreateGC(wp);

  XMapWindow(wp->display, wp->toplevel_window);
  XMapSubwindows(wp->display, wp->toplevel_window);
  XFlush(wp->display);
}

/*****************************************************************************/
/* Initializing Viewer.                                                        */
/*****************************************************************************/

static void InitializeViewer(VIEWER * wp)
{
  memset(&wp->status, 0, sizeof(wp->status));
}

static unsigned long int DispGetColor(VIEWER * wp, char * color_name)
{
  XColor c0, c1;
  XAllocNamedColor(wp->display, wp->colormap, color_name, &c0, &c1);
  return (c0.pixel);
}

static void InitializeGC(VIEWER * wp)
{
  wp->gc_line       = XCreateGC(wp->display, wp->canvas_window, 0, 0);
  wp->gc_line_clear = XCreateGC(wp->display, wp->canvas_window, 0, 0);
  wp->gc_line_green = XCreateGC(wp->display, wp->canvas_window, 0, 0);
  wp->gc_line_red   = XCreateGC(wp->display, wp->canvas_window, 0, 0);

  XSetForeground(wp->display, wp->gc_line      , wp->foreground_color);
  XSetForeground(wp->display, wp->gc_line_clear, 0);
  XSetForeground(wp->display, wp->gc_line_green, DispGetColor(wp, "green"));
  XSetForeground(wp->display, wp->gc_line_red  , DispGetColor(wp, "red"));

  XSetFunction(wp->display, wp->gc_line_clear, GXcopy);
  XSetFunction(wp->display, wp->gc_line_green, GXxor);
  XSetFunction(wp->display, wp->gc_line_red  , GXxor);

  XSetLineAttributes(wp->display, wp->gc_line,
		     wp->thickness, LineSolid, CapButt, JoinMiter);
  XSetLineAttributes(wp->display, wp->gc_line_clear,
		     wp->thickness, LineSolid, CapButt, JoinMiter);
  XSetLineAttributes(wp->display, wp->gc_line_green,
		     wp->thickness, LineSolid, CapButt, JoinMiter);
  XSetLineAttributes(wp->display, wp->gc_line_red,
		     wp->thickness, LineSolid, CapButt, JoinMiter);
}

/*****************************************************************************/
/* Quit                                                                      */
/*****************************************************************************/

static void Quit()
{
  exit (0);
}

/*****************************************************************************/
/* Addition of EventHandler                                                  */
/*****************************************************************************/

static void DrawCanvas(VIEWER * wp);

static void Event_Pointer(Widget w, XtPointer p, XEvent * event, Boolean * dispatch)
{
  VIEWER * wp;
  XPoint pointer;

  wp = (VIEWER *)p;
  pointer.x = event->xmotion.x;
  pointer.y = event->xmotion.y;

  /* none */
}

static void Event_Exposure(Widget w, XtPointer p, XEvent * event, Boolean * dispatch)
{
  VIEWER * wp;
  wp = (VIEWER *)p;
  if (event->type == Expose) {
    DrawCanvas(wp);
  }
  fprintf(stderr, "expose\n");
}

static void Event_Key(Widget w, XtPointer p, XEvent * event, Boolean * dispatch)
{
  VIEWER * wp;
  KeySym symbol;

  wp = (VIEWER *)p;
  if (event->type == KeyPress) {
    symbol = XKeycodeToKeysym(event->xkey.display, event->xkey.keycode, 0);
    if (symbol == XK_q) Quit(wp);
    if (symbol == XK_c) InitializeViewer(wp);
    DrawCanvas(wp);
  }
}

static void Event_Button(Widget w, XtPointer p, XEvent * event, Boolean * dispatch)
{
  VIEWER * wp;
  wp = (VIEWER *)p;
  if (event->type == ButtonPress) {
    if (w == wp->ls_come) wp->status.ls_come = !wp->status.ls_come;
    if (w == wp->ls_pass) wp->status.ls_pass = !wp->status.ls_pass;
    if (w == wp->rs_come) wp->status.rs_come = !wp->status.rs_come;
    if (w == wp->rs_pass) wp->status.rs_pass = !wp->status.rs_pass;

    if (w == wp->intr) wp->status.intr = !wp->status.intr;
    if (w == wp->init_button) InitializeViewer(wp);
    if (w == wp->quit_button) Quit(wp);

    DrawCanvas(wp);
  }
}

static void AddEvent(VIEWER * wp)
{
  XtAddEventHandler(wp->toplevel, PointerMotionMask, False, Event_Pointer, wp);
  XtAddEventHandler(wp->toplevel, ExposureMask, False, Event_Exposure, wp);
  XtAddEventHandler(wp->toplevel, KeyPressMask, False, Event_Key, wp);

  XtAddEventHandler(wp->ls_come,   ButtonPressMask, False, Event_Button, wp);
  XtAddEventHandler(wp->ls_pass,   ButtonPressMask, False, Event_Button, wp);
  XtAddEventHandler(wp->rs_come,   ButtonPressMask, False, Event_Button, wp);
  XtAddEventHandler(wp->rs_pass,   ButtonPressMask, False, Event_Button, wp);

  XtAddEventHandler(wp->intr,        ButtonPressMask, False, Event_Button, wp);
  XtAddEventHandler(wp->init_button, ButtonPressMask, False, Event_Button, wp);
  XtAddEventHandler(wp->quit_button, ButtonPressMask, False, Event_Button, wp);
}

/*****************************************************************************/
/* Setting Window Attributes.                                                */
/*****************************************************************************/

static void SetWindowAttributes(VIEWER * wp)
{
  XSizeHints sh;
  XWMHints wmhints;

  sh.flags = (PBaseSize | PResizeInc | PMinSize | PMaxSize);
  sh.width_inc = sh.height_inc = 0;
  sh.base_width  = sh.min_width  = sh.max_width  = wp->width;
  sh.base_height = sh.min_height = sh.max_height = wp->height;
  XSetWMNormalHints(wp->display, wp->toplevel_window, &sh);

  wmhints.input = True;
  XSetWMHints(wp->display, wp->toplevel_window, &wmhints);
  XSetWMHints(wp->display, wp->canvas_window, &wmhints);
  wp->atom_1 = XInternAtom(wp->display, "WM_PROTOCOLS", False);
  wp->atom_2 = XInternAtom(wp->display, "WM_DELETE_WINDOW", False);
  XSetWMProtocols(wp->display, wp->toplevel_window, &(wp->atom_2), 1);

  XStoreName(wp->display, wp->toplevel_window, "Viewer");
}

/*****************************************************************************/
/* Viewer.                                                                     */
/*****************************************************************************/

static void ClearCanvas(VIEWER * wp)
{
  XFillRectangle(wp->display, wp->canvas_pixmap, wp->gc_background, 0, 0,
                 wp->width, wp->height);
}

static void Handler(VIEWER * wp)
{
  ;
}

static void DrawPlane(VIEWER * wp)
{
  XDrawLine(wp->display, wp->canvas_pixmap, wp->gc_line,
	    50, 280, 350, 280 - (90 - wp->params.bar_angle) * 2);
}

static void DrawCanvas(VIEWER * wp)
{
  ClearCanvas(wp);
  Handler(wp);
  DrawPlane(wp);

  XCopyArea(wp->display, wp->canvas_pixmap, wp->canvas_window, wp->gc_copy,
	    0, 0, wp->width, wp->height, 0, 0);

  DrawPanel(wp);
  XFlush(wp->display);
}

/*****************************************************************************/
/* Main Loop.                                                                */
/*****************************************************************************/

#define EVERYDRAW
#define USE_TIMEOUT

#ifdef EVERYDRAW
int serial_tstc(int fd)
{
  fd_set fds;
  struct timeval timeout;
  int ret;

  if (fd == 0)
    return 0;

  FD_ZERO(&fds);
  FD_SET(fd, &fds);

  timeout.tv_sec = 0;
  timeout.tv_usec = 0;

  ret = select(fd + 1, &fds, NULL, NULL, &timeout);
  if (ret > 0) {
    if (FD_ISSET(fd, &fds)) {
      return 1;
    }
  }

  return 0;
}

int serial_getc(int fd)
{
  char c;
  if (!serial_tstc(fd))
    return '\0';
  read(fd, &c, 1);
  return c;
}

char *serial_gets(int fd)
{
  static char buf[32];
  static char *p = buf;
  int c;

  while (1) {
    c = serial_getc(fd);
    if (c == '\0')
      break;

    if ((c == '\n') || (c == '\r')) {
      if (p != buf) {
	*p = '\0';
	p = buf;
	return buf;
      }
      return NULL;
    }
    *(p++) = c;
  }

  return NULL;
}

int serial_puts(int fd, char *str)
{
  char buf[32];
  printf("%s\n", str);
  fflush(stdout);
  sprintf(buf, "%s\n", str);
  write(fd, buf, strlen(buf));
}

int cmdexec(char *cmd, VIEWER * wp)
{
  printf("%s\n", cmd);
  fflush(stdout);

  /* ѥ᡼ */
  if (!strcmp(cmd, "LS_COME")) {
    serial_puts(wp->params.fd, wp->status.ls_come ? "ON" : "OFF");
    return 0;
  }
  if (!strcmp(cmd, "LS_PASS")) {
    serial_puts(wp->params.fd, wp->status.ls_pass ? "ON" : "OFF");
    return 0;
  }
  if (!strcmp(cmd, "RS_COME")) {
    serial_puts(wp->params.fd, wp->status.rs_come ? "ON" : "OFF");
    return 0;
  }
  if (!strcmp(cmd, "RS_PASS")) {
    serial_puts(wp->params.fd, wp->status.rs_pass ? "ON" : "OFF");
    return 0;
  }
  if (!strcmp(cmd, "US_PUSH")) {
    serial_puts(wp->params.fd, wp->params.upsw ? "ON" : "OFF");
    return 0;
  }
  if (!strcmp(cmd, "DS_PUSH")) {
    serial_puts(wp->params.fd, wp->params.downsw ? "ON" : "OFF");
    return 0;
  }

  /*  */
  if (!strcmp(cmd, "OPEN")) {
    wp->params.bar_angle_delta = -2;
    serial_puts(wp->params.fd, "OK");
    return 0;
  }
  if (!strcmp(cmd, "CLOSE")) {
    wp->params.bar_angle_delta =  2;
    serial_puts(wp->params.fd, "OK");
    return 0;
  }
  if (!strcmp(cmd, "STOP")) {
    wp->params.bar_angle_delta =  0;
    serial_puts(wp->params.fd, "OK");
    return 0;
  }
  if (!strcmp(cmd, "LIGHTON")) {
    wp->params.light = 1;
    serial_puts(wp->params.fd, "OK");
    return 0;
  }
  if (!strcmp(cmd, "LIGHTOFF")) {
    wp->params.light = 0;
    serial_puts(wp->params.fd, "OK");
    return 0;
  }

  serial_puts(wp->params.fd, "UNKNOWN");
  return 0;
}

int read_serial(VIEWER * wp)
{
  char *cmd;

  while (1) {
    cmd =  serial_gets(wp->params.fd);
    if (!cmd)
      return 0;
    cmdexec(cmd, wp);
  }
}

static void MainViewer(VIEWER * wp)
{
  static int init = 0;
  static int ls_come, ls_pass, rs_come, rs_pass, upsw, downsw;

  if (init && wp->status.intr) {
    if ((!ls_come && wp->status.ls_come) ||
	(!ls_pass && wp->status.ls_pass) ||
	(!rs_come && wp->status.rs_come) ||
	(!rs_pass && wp->status.rs_pass) ||
	(!upsw    && wp->params.upsw) ||
	(!downsw  && wp->params.downsw)) {
      printf("INTR\n");
      fflush(stdout);
      serial_puts(wp->params.fd, "INTR");
    }
  }

  ls_come = wp->status.ls_come;
  ls_pass = wp->status.ls_pass;
  rs_come = wp->status.rs_come;
  rs_pass = wp->status.rs_pass;
  upsw    = wp->params.upsw;
  downsw  = wp->params.downsw;
  init = 1;

  read_serial(wp);

  wp->params.bar_angle += wp->params.bar_angle_delta;

  wp->params.upsw = 0;
  wp->params.downsw = 0;
  wp->params.limit = 0;
  if (wp->params.bar_angle < 0) {
    wp->params.bar_angle = 0;
    wp->params.limit = 1;
  }
  if (wp->params.bar_angle > 90) {
    wp->params.bar_angle = 90;
    wp->params.limit = 1;
  }
  if (wp->params.bar_angle ==  0)
    wp->params.upsw = 1;
  if (wp->params.bar_angle == 90)
    wp->params.downsw = 1;

  DrawCanvas(wp);
}

#ifdef USE_TIMEOUT
static void TimeOutViewer(XtPointer p, XtIntervalId *id)
{
  VIEWER * wp;
  wp = (VIEWER *)p;
  MainViewer(wp);
  XtAppAddTimeOut(wp->app_context, wp->wait / 1000, TimeOutViewer, wp);
}
#else
static Boolean WorkProcViewer(XtPointer p)
{
  VIEWER * wp;
  wp = (VIEWER *)p;
  MainViewer(wp);
  usleep(wp->wait);
  return (False);
}
#endif
#endif

static int setterm(int fd)
{
#if 1
  struct termios term;

  tcgetattr(fd, &term);
  term.c_lflag &= ~ECHO;
  term.c_lflag &= ~ICANON;
  term.c_cc[VTIME] = 0;
  term.c_cc[VMIN] = 1;
  tcsetattr(fd, TCSANOW, &term);
#endif

  return 0;
}

static void InitializeParams(VIEWER * wp, int argc, char *argv[])
{
  int fd;
  memset(&wp->params, 0, sizeof(wp->params));
  fd = open(argv[1], O_RDWR | O_NONBLOCK | O_NOCTTY);
  if (fd < 0) {
    fprintf(stderr, "serial open error %s\n", argv[1]);
    exit(1);
  }
  fprintf(stderr, "fd %d\n", fd);
  wp->params.fd = fd;

  setterm(fd);
}

int main(int argc, char *argv[])
{
  VIEWER viewer;

  InitializeValues(&viewer);
  InitializeParams(&viewer, argc, argv);
  InitializeDisplay(&viewer, argc, argv);
  InitializeViewer(&viewer);
  InitializeGC(&viewer);
  AddEvent(&viewer);
  SetWindowAttributes(&viewer);

#ifdef EVERYDRAW
#ifdef USE_TIMEOUT
  TimeOutViewer((XtPointer)&viewer, NULL);
#else
  XtAppAddWorkProc(viewer.app_context, WorkProcViewer, &viewer);
#endif
#endif

  XtAppMainLoop(viewer.app_context);

  return 0;
}

/*****************************************************************************/
/* End of File.                                                              */
/*****************************************************************************/
